use clichess::{ClientRequest, GameInfo, RecvPositionError, UserRole, EXIT_MSG};
use shakmaty::{Chess, Color, Outcome, Position, Setup};
use std::io;
use std::os::unix::net::UnixStream;
use std::sync::mpsc::{channel, Receiver, TryRecvError};
use std::thread;
use std::time;

struct Client {
    player: clichess::Player,
    side: Color,
    keyboard_input_recv: Receiver<String>,
    server_message_recv: Receiver<String>,
    opponent_name: Option<String>,
    last_move: Option<String>,
}

fn main() {
    let version = env!("CARGO_PKG_VERSION");
    println!("Running clichess version {}", version);
    let username = std::env::args().nth(1).expect("no name given");
    let public_key = std::env::args().nth(2).expect("no public key given");
    //send username and public key to server
    let mut stream = match UnixStream::connect("/tmp/clichess.socket") {
        Ok(sock) => sock,
        Err(_) => {
            println!("clichess daemon is not running.");
            return;
        }
    };
    let keyboard_input_recv = setup_keyboard_input_recv();
    //start a thread to listen to server messages
    let server_message_recv = setup_server_message_recv(&stream).unwrap();

    let mut client = Client {
        player: clichess::Player {
            role: UserRole::Spectator,
            username,
            public_key,
        },
        side: Color::White,
        keyboard_input_recv,
        server_message_recv,
        opponent_name: Option::None,
        last_move: Option::None,
    };

    let login_result = login(&mut client, &mut stream);
    if login_result == LoginResult::UserExited {
        send_request(&client, &mut stream, ClientRequest::Exit);
        return;
    }
    if client.player.role == UserRole::Black {
        client.side = Color::Black;
    }
    if client.player.role == UserRole::Spectator {
        println!(
            "Hello, {} !\n\r You're spectating !",
            client.player.username
        )
    } else {
        println!(
            "Hello, {} !\n\r You're playing with the {} pieces",
            client.player.username,
            client.player.role.to_string()
        );
    }
    let mut current_position = fetch_initial_chess_position(&client, &mut stream);
    loop {
        let is_player_turn = clichess::is_player_turn(&client.player, current_position.turn());
        if is_player_turn || client.player.role == UserRole::Spectator {
            client
                .opponent_name
                .clone()
                .map(|name| print!("{} played", name));
            client
                .last_move
                .clone()
                .map(|chessmove| println!(" {}", chessmove));
        }
        println!(
            "{}",
            clichess::board_representation(&current_position, client.side)
        );
        //check if game is over.
        if current_position.is_game_over() {
            break;
        }
        if is_player_turn {
            //it's the user turn, taking user input
            let mut input = client.keyboard_input_recv.recv().unwrap();
            input = String::from(input.trim());
            if input == EXIT_MSG {
                break;
            }
            let response = send_request(&client, &mut stream, ClientRequest::Play(input.clone()));
            if response == String::from("KO") {
                println!("Invalid move: {}", input);
                //go back to taking user input
                continue;
            }
            client.last_move = Some(input);
            match get_current_position(&mut client, &mut stream) {
                Ok(position) => current_position = position,
                Err(_) => break,
            };
        } else {
            match wait_for_next_move(&mut client, &mut stream) {
                Ok(position) => current_position = position,
                Err(_) => break,
            };
        }
    }
    match current_position.outcome() {
        None => println!("Bye"),
        Some(Outcome::Draw) => println!("Draw game."),
        Some(Outcome::Decisive { winner }) => {
            if winner == Color::White {
                println!("White has won the game.")
            } else {
                println!("Black has won the game.")
            }
        }
    }
    send_request(&client, &mut stream, ClientRequest::Exit);
}

#[derive(PartialEq)]
enum LoginResult {
    Success,
    UserExited,
}

fn login(client: &mut Client, stream: &mut UnixStream) -> LoginResult {
    let mut response = String::from("KO");
    while &response != "OK" {
        match prompt_user_for_role(client, stream) {
            PromptedRoleResponse::Role(role) => {
                client.player.role = role.clone();
                response = send_request(
                    client,
                    stream,
                    ClientRequest::Login {
                        username: client.player.username.clone(),
                        pubkey: client.player.public_key.clone(),
                        role,
                    },
                );
            }
            PromptedRoleResponse::Exit => return LoginResult::UserExited,
            PromptedRoleResponse::Retry => continue,
        };
    }
    LoginResult::Success
}

fn send_request(client: &Client, stream: &mut UnixStream, request: ClientRequest) -> String {
    let response: String;
    match serde_json::to_string(&request) {
        Ok(request_str) => {
            clichess::write_to_stream(stream, request_str).unwrap();
            response = fetch_message_from_server(client);
        }

        Err(e) => {
            println!("Error when parsing client request: {}", e);
            response = String::from("KO");
        }
    };
    response
}

fn setup_keyboard_input_recv() -> Receiver<String> {
    let (sender, receiver) = channel();
    thread::spawn(move || {
        loop {
            let mut buffer = String::new();
            //wait for user input
            io::stdin().read_line(&mut buffer).unwrap();
            sender.send(buffer).unwrap();
        }
    });
    receiver
}

fn setup_server_message_recv(stream: &UnixStream) -> io::Result<Receiver<String>> {
    let (sender, receiver) = channel();
    let thread_stream = stream.try_clone()?;

    thread::spawn(move || {
        loop {
            //wait for server message
            let buffer =
                clichess::read_line_from_stream(&thread_stream).expect("Error message from server");
            sender.send(buffer).unwrap();
        }
    });
    Ok(receiver)
}

//get current position from server
fn get_current_position(
    client: &mut Client,
    stream: &mut UnixStream,
) -> Result<Chess, RecvPositionError> {
    let response = send_request(client, stream, ClientRequest::GetGameInfo);
    if response.is_empty() {
        Err(RecvPositionError::UserCanceledError)
    } else {
        let game_info: GameInfo = serde_json::from_str(&response).unwrap();
        update_client(&game_info, client);
        Ok(clichess::parse_position(&game_info.game_fen))
    }
}

fn wait_for_next_move(
    client: &mut Client,
    stream: &mut UnixStream,
) -> Result<Chess, RecvPositionError> {
    println!("Waiting for opponent move...");
    send_request(client, stream, ClientRequest::WaitForNextMove);
    let response;
    loop {
        thread::sleep(time::Duration::from_millis(10));
        {
            let server_response_recv = &client.server_message_recv;
            match server_response_recv.try_recv() {
                Ok(msg) => {
                    response = msg;
                    break;
                }
                Err(TryRecvError::Disconnected) => println!("Error: server disconnected."),
                Err(TryRecvError::Empty) => continue,
            }
        }
    }
    if response.is_empty() {
        Err(RecvPositionError::UserCanceledError)
    } else {
        let game_info: GameInfo = serde_json::from_str(&response).unwrap();
        update_client(&game_info, client);
        Ok(clichess::parse_position(&game_info.game_fen))
    }
}

fn update_client(game_info: &GameInfo, client: &mut Client) {
    if game_info.opponent_name != "" {
        client.opponent_name = Some(game_info.opponent_name.clone());
    }
    if game_info.last_move != "" {
        client.last_move = Some(game_info.last_move.clone());
    }
}

fn fetch_message_from_server(client: &Client) -> String {
    let response;
    loop {
        thread::sleep(time::Duration::from_millis(10));
        {
            let server_response_recv = &client.server_message_recv;
            match server_response_recv.try_recv() {
                Ok(msg) => {
                    response = msg;
                    break;
                }
                Err(TryRecvError::Disconnected) => println!("Error: server disconnected."),
                Err(TryRecvError::Empty) => continue,
            }
        }
    }
    response
}

enum PromptedRoleResponse {
    Role(UserRole),
    Exit,
    Retry,
}

fn prompt_user_for_role(client: &Client, stream: &mut UnixStream) -> PromptedRoleResponse {
    let response = send_request(client, stream, ClientRequest::FetchAvailableRoles);
    let available_roles = roles_from_str(&response).unwrap();
    let mut prompt = String::new();
    if !available_roles.contains(&UserRole::White) && !available_roles.contains(&UserRole::Black) {
        prompt.push_str("You can only spectate this game. Press enter to start spectating.");
    } else if available_roles.contains(&UserRole::White) {
        prompt = String::from("Do you want to play as White (W)");
        if available_roles.contains(&UserRole::Black) {
            prompt.push_str(", Black (B)");
        }
        prompt.push_str(" or Spectate (S)?");
    } else {
        prompt = String::from("Do you want to play as Black (B) or Spectate (S)?");
    }
    println!("{}", prompt);
    let mut input = String::from(client.keyboard_input_recv.recv().unwrap().trim());
    if input.trim() == EXIT_MSG {
        return PromptedRoleResponse::Exit;
    }

    if !available_roles.contains(&UserRole::White) && !available_roles.contains(&UserRole::Black) {
        //we can only spectate
        input = String::from("S");
    }
    match clichess::parse_to_role(&input) {
        Ok(r) => {
            if !available_roles.contains(&r) {
                println!("Sorry, this side is not available.");
                return PromptedRoleResponse::Retry;
            }
            return PromptedRoleResponse::Role(r);
        }
        Err(e) => {
            println!("{}", e);
            return PromptedRoleResponse::Retry;
        }
    };
}

fn fetch_initial_chess_position(client: &Client, stream: &mut UnixStream) -> Chess {
    let response = send_request(client, stream, ClientRequest::GetGameInfo);
    let game_info: GameInfo = serde_json::from_str(&response).unwrap();
    clichess::parse_position(&game_info.game_fen)
}

fn roles_from_str(s: &str) -> Result<Vec<UserRole>, String> {
    s.split(',').map(clichess::parse_to_role).collect()
}
